! This file contains: (ordered to all can be in one file)
!   (1) ErrorCalcs module for error calcs,
!   (2) BVPmod main solution module
!   (3) a small main
!-------------------------------------------------------------------------------
Module ErrorCalcs
!-------------------------------------------------------------------------------
! This module is an add on to calculate exact solution and various error norms
! Results use analytical solution when indicated or the variable coefficient
! problem in the text. For the variable coefficient problem errors are calculated
! with n = 18 Galerkin method.
! ------------------------------------------------------------------------------
Use Array_Print, only : ArrayPrint, VectorPrint
Use OrthogonalColloc, Only :  ColDataType,Initialize,Xpoints,WeightQ,Interp,Lcoeff, &
      Nonsymmetric,Gauss,Lobatto,Chebyshev
Use Orth_Poly, only : Jacobi_Transform
include 'defs.fi'
Private
Public ErrCalc
Public Exact
Public ErrorNorm
Public ErrorNorm1
Public ErrResid
Public ErrTransform
Real(float), private, parameter :: Debug = 0
integer, public, parameter :: nsolv = 4
Logical :: ConstCoef,Dirichlet
Real(float) :: Thparm, Biot(2)
Type(ColDataType) :: OX
! ----------------------------------------------------------
Integer, public, parameter :: Nex = 20  ! Aparm = 0.2, Dirichlet, Galerkin solution, n=18
Real(float), public :: Xex(Nex),Wex(Nex)
Real(float), parameter :: Yex(Nex) = (/  0.0D0, &  ! Galerkin n=18, Aparm = 0.2
 0.0478219799410479D0,0.1523777644219207D0,0.2966502067945801D0,0.4595608167038157D0,0.6192400924032583D0,0.7566744526443667D0, &
 0.8599957579486170D0,0.9272444163606441D0,0.9647356106898428D0,0.9817545988621290D0,0.9852514371620389D0,0.9766200914995425D0, &
 0.9496909742032595D0,0.8892596717337564D0,0.7735016761492272D0,0.5873659266109881D0,0.3477644519776378D0,0.1207109007847567D0, &
 0.0D0/)
Real(float), parameter :: Xfluxex(2) = (/0.05061659803731418D0,0.13367790943869215D0/)
!-------------------------------------------------------------------------------
Contains
!-------------------------------------------------------------------------------
Subroutine ErrCalc(Const,DirBC,Thp,Bi)
   Logical :: Const,DirBC
   Real(float) :: Thp,Bi(2)
   ConstCoef = Const
   Dirichlet = DirBC
   Thparm = Thp
   Biot = Bi
   Call Initialize(OX,Nex-2,Lobatto)
   Xex = Xpoints(OX)
   Wex = WeightQ(OX)
End Subroutine ErrCalc
!-------------------------------------------------------------------------------
Function Exact(x,xflux)  result(y)
   Real(float) :: x(:),y(size(x)),xx(size(x))
   Real(float),optional :: xflux(:)
   Real(float) :: cmult,xeff
   ! calculates "exact" solution for constant coefficients or for the
   ! variable coefficient case in the text (assumes Galerkin n=18 is exact)
!-------------------------------------------------------------------------------
   if(ConstCoef)then
      xeff = tanh(Thparm)/Thparm
      if(Dirichlet)then
         cmult = 1.0_float/cosh(Thparm)
      else
         cmult = 1.0_float/(cosh(Thparm) + (Thparm/Biot(1))*sinh(Thparm))
         xeff = 1.0_float/(1.0_float/xeff + Thparm*Thparm/Biot(1))
      endif
      xx = Thparm*(2.0_float*x - 1.0_float)
      y = 1.0_float - cmult*cosh(xx)
      if(present(xflux))then
         xflux(1) = xeff;  xflux(2:3) = xeff*0.5_float
      endif
   else
      y = Interp(x,OX,Yex)   ! "exact" solution at x
      if(present(xflux))then
         xflux(2:3) = Xfluxex;  xflux(1) = sum(Xfluxex)
      endif
   endif
End Function Exact
!-------------------------------------------------------------------------------
Function ErrorNorm(OC,y) result(errn)
   Real(float) :: errn(nsolv),y(:,:)
   Type(ColDataType) :: OC
   Real(float) :: yx(Nex,nsolv),yexact(Nex)
   integer :: j
   ! --------------------------------------
   ! Calculate L2 error norm with n=18 (interior) Lobatto quadrature
   if(ConstCoef)then
      yexact = Exact(Xex)  ! const. coefficients, analytic solution
   else
      yexact = Yex         ! n=18, Galerkin ~exact, Aparm = 0.2, Dirichlet
   endif
   yx = Interp(Xex,OC,y)   ! interpolating solutions onto n=18 Lobatto
   do j=1,nsolv
      errn(j) = Sum(Wex(:)*(yexact(:) - yx(:,j))**2)
      errn(j) = sqrt(errn(j))
   enddo
End Function ErrorNorm
!-------------------------------------------------------------------------------
Function ErrorNorm1(OC,y) result(errn)
   Real(float) :: errn(nsolv,2),y(:,:)
   Type(ColDataType) :: OC
   integer, parameter :: Nv=200
   Real(float), parameter :: delx = 1.0/Real(Nv,float)
   integer :: i,j
   Real(float), Parameter :: xv(0:Nv) = (/(i*delx,i=0,Nv)/)
   Real(float), Parameter :: wv(0:Nv) = (/delx*0.5_float,(delx,i=1,Nv-1),delx*0.5_float/)
   Real(float) :: yx(0:Nv,nsolv),yexact(0:Nv)
   ! ---------------------------------------
   ! Calculate L1 & L2 error norms with Nv + 1 point trapezoidal rule
   yexact = Exact(xv)
   yx = Interp(xv,OC,y)    ! interpolating solutions onto trapezoidal pts.
   do j=1,nsolv
      errn(j,1) = Sum(Wv(:)*(yexact(:) - yx(:,j))**2)
      errn(j,1) = sqrt(errn(j,1))
      errn(j,2) = Sum(Wv(:)*abs(yexact(:) - yx(:,j)))
   enddo
End Function ErrorNorm1
!-------------------------------------------------------------------------------
function ErrResid(r,t) result(rerr)
   real(float) :: r(0:),rerr(2),xx(Nex)
   integer :: t
   xx = 1.0_float
   if(t > Gauss)xx = xex*(1.0_float-xex)
   if(t == Chebyshev)xx = sqrt(xx)
   rerr(1) = sqrt(sum(Wex*r*r))  ! L2 error
   rerr(2) = sqrt(sum(Wex*xx*r*r)/sum(Wex*xx))  ! Lw weighted error
   !write(iout,'(a,2g14.7)')'ErrResid:',sum(Wex*xx*r*r),sum(Wex*xx)
end function ErrResid
!-------------------------------------------------------------------------------
function ErrTransform(na,x,R,t) result(an)
   integer :: na,t,nx
   real(float) :: x(0:),R(0:),an(0:na)
   integer, parameter :: g = Nonsymmetric, s = Nonsymmetric
   nx = size(x) - 2
   an = Jacobi_Transform(na,t,g,Func=Rescalc,Shift=.true.)   ! calculate transform
   contains
   ! ----------------------------------------
   function ResCalc(xx)  result(rx)
      real(float), intent(in) :: xx(:)
      real(float) :: rx(size(xx)),Lc(size(xx),nx+2)
      Lc = Lcoeff(xx,x)
      rx = MatMul(Lc,R)
   end function ResCalc
End function ErrTransform
!-------------------------------------------------------------------------------
End Module ErrorCalcs
!-------------------------------------------------------------------------------
!  Module BVPmod (just below) solves the problem
!  It solves a nonsymmetric reaction & diffusion problem using
!  Orthogonal Collocation with a choice of collocation points
!  and with Moments or Galerkin methods with various options
!  for integration of the rate terms.
!  Solves for Planar Geometry with no consideration of symmetry
!  For planar geometry the equations are:
!     d2y/dx2 + c(x)*(1-y) = 0
!  with y = 0 at x = 0,1
!  The effectiveness factor is calculated by three different
!  methods:
!     1. By integration to calculate the average reaction rate
!     2. Using the derivative of the solution at x = 0 & 1
!     3. Using the flux relation which corresponds to a natural
!        treatment of a flux boundary condition
!  c(x) = 4*q(x)*phi^2, phi is the Thiele Modulus (5 used in most examples)
!  q(x) is a cubic function with average value of 1, its value varies from
!  q(0) = Aparm, q(1) = 2 - Aparm, q(0.5) = 1
!  Aparm = .2 gives 0.2 at left, 1.8 at right (case in text)
!  Aparm = 1, gives no variation
! ----------------------------------------------------------------
Module BVPmod
   Use OrthogonalColloc, Only : ColDataType,Initialize,CollocCoef,MassMatrix,Lcoeff,  &  !PolyTransform,  &
         Chebyshev,TxtType
   Use ErrorCalcs, only : ErrCalc, Exact, ErrorNorm, ErrorNorm1, ErrTransform, ErrResid, Nex, Xex
   Use Array_Print, only : ArrayPrint, VectorPrint
   include 'defs.fi'
   Private
   Public SolveBVP
   Public ReportErr
   Real(float), private, parameter :: Debug = 0
   integer, public, parameter :: nsolv = 4, nflux = 6
   integer :: ityp
   Real(float) :: Thparm,Thparm2             ! Thiele parameter & 4*Thiele^2 (constant in source)
   Real(float) :: Aparm, Aparm1              ! spatial variation parameters in source term
   Real(float) :: Biot(0:1),Biot2(0:1)       ! left & right side Biot numbers in flux condition
   Logical :: Dirichlet,ConstCoef,Symmetric  ! logicals Dirichlet B.C.?, Constant Coefficients?
   Type(ColDataType) :: OC
   Integer, Parameter :: Nx = Nex ! too override these with Xex from Error routine
!-------------------------------------------------------------------------------
   Contains
!-------------------------------------------------------------------------------
Subroutine SolveBVP(n,itypx,Thx,Apx,bcfx,errfx,errnm)
   Use LUSolvers, Only : LUSolveC, LUSolveSQ
   Integer, intent(in) :: n      ! interior points
   Integer :: itypx               ! Gauss = 1, Lobatto = 2
   real(float) :: Thx            ! average Thiele parameter
   Real(float) :: Apx            ! left side reactivity, right is 2 - Aparm
   Real(float) :: bcfx(2)        ! left and right side Biot number in flux condition
   Real(float) :: errfx(:,:)     ! flux errors
   Real(float) :: errnm(:,:)     ! profile error norms
   Real(float) :: flux(0:nflux,nsolv)     ! flux
   Real(float) :: Ac(0:n+1,0:n+1),Bc(0:n+1,0:n+1),Cc(0:n+1,0:n+1),Dc(0:n+1,0:n+1)
   Real(float) :: xc(0:n+1),wc(0:n+1),yc(0:n+1,nsolv),rc(0:n+1),rs(0:n+1)
   Real(float) :: Ax(0:n+1,0:n+1),Af(0:n+1,0:n+1)
   Integer :: i,n1,i0,i1,k,mx
   character*(18) :: txtmass
   ! ---------------------------------------------------------------------------
   ! This program solves diffusion/conduction with linear source term.  The source
   ! term may have a variable coefficient as determined by parameter Apx.
   ! The problem is solved by 4 different methods:
   !  1. conventional orthogonal collocation (pseudospectral)
   !  2. equivalent formulation with symmetric matric, corrected B.C. treatement
   !  3. Moments (Gauss) or Galerkin (Lobatto) with interpolated souce terms
   !  4. Moments (Gauss) or Galerkin (Lobatto) with exact integration
   ! ---------------------------------------------------------------------------
   !  Problem parameters
   n1 = n+1
   ityp = itypx
   Thparm = Thx
   Thparm2 = (Thx*2.0_float)**2        ! 2.0 to account for full slab
   Aparm = Apx; Aparm1 = (1.0_float - Aparm)*4.0_float
   Aparm = Thparm2*Aparm;  Aparm1 = Thparm2*Aparm1
   Biot = bcfx                         ! left & right BC Biot (use 0 for Dirichlet)
   Biot2 = Biot*2.0_float              ! 2.0 to account for full slab
   ConstCoef = abs(Aparm1) < 1.0e-8    ! Constant coefficients
   Dirichlet = maxval(Biot) < 1.0e-8   ! Dirichlet or 3rd kind
   Symmetric = .not.(ityp == Chebyshev)
   if(Dirichlet)then
   ! for Dirichlet conditions, eqns are set up at 0 and n+1, but not used
      i0 = 1;  i1 = n   ! Dirichlet y(0) = y(n+1) = 0
   else
      i0 = 0;  i1 = n1  ! 3rd kind, solve for y(0) & y(n+1)
   endif
   !-------  Get the collocation data
   Call Initialize(OC,n,ityp)
   Call CollocCoef(OC,xc,wc,Ac,Bc,Cc)
   rc = FuncRX(xc)
   Call ReportBasic(xc,wc,rc,Ac,Bc,Cc,ityp)
   ! ------- Create Solution Matrix & Solve, 4 different methods
   yc(0,:) = 0.0_float; yc(n1,:) = 0.0_float
   ! 1. ------ Conventional orthogonal collocation -----------------------------
   ! use first derivative -Ac(0)*y = Bi*y(0), etc
   do i=0,n+1
      Ax(i,:) = Bc(i,:)
      Ax(i,i) = Ax(i,i) - rc(i)
   enddo
   rs = -rc
   Ax(0,:) = -Ac(0,:);  Ax(0,0) = Ax(0,0) + Biot2(0);     rs(0) = 0
   Ax(n1,:) = Ac(n1,:); Ax(n1,n1) = Ax(n1,n1) + Biot2(1); rs(n1) = 0
   yc(i0:i1,1) = LUSolveC(Ax(i0:i1,i0:i1),rs(i0:i1))     ! Use Nonsymmetric Solver
   flux(:,1) = FluxCalc(rc,wc,Ac,Bc,Ax,rs,yc(:,1),1)
   call ArrayPrint("Matrix & RHS, conventional:",Ax(i0:i1,i0:i1),rs(i0:i1))
   if(Debug > 0)Call VectorPrint("Flux",flux(:,1))
   ! ------ Solve with 3 method differing only by mass matrix
   !  2. collocation, diagonal mass matrix
   !  3. interpolation of rates, full mass matrix
   !  4. integration by quadrature, m = n + mx quadrature points
   mx = 2   ! 2 gives exact, 1 approx., 0 collocation
   do k=2,4
      select case(k)
      case(2)
         txtmass = 'collocation (weak)'
         Dc = 0.0_float
         rs = rc*Wc
         forall(i=0:n+1)Dc(i,i) = rs(i)
      case(3)
         txtmass = 'interpolated rates'
         Dc = MassMatrix(OC,rc)  ! interpolated mass matrix
      case(4)
         txtmass = 'exact integration'
         Dc = MassMatrix(OC,FuncRX,mx) ! full mass matrix mx extra points
      end select
      if(Debug > 0)call ArrayPrint("Mass Matrix, "//txtmass,Dc)
      forall(i=0:n+1)rs(i) = sum(Dc(i,:))
      Ax = Cc + Dc
      Af = Ax; Ax(0,0) = Ax(0,0) + Biot2(0);  Ax(n1,n1) = Ax(n1,n1) + Biot2(1)
      yc(i0:i1,k) = LUSolveC(Ax(i0:i1,i0:i1),rs(i0:i1))  ! symmetric Galerkin only
      flux(:,k) = FluxCalc(rc,wc,Ac,Bc,Af,rs,yc(:,k),k)
      call ArrayPrint("Matrix & RHS, "//txtmass,Ax(i0:i1,i0:i1),rs(i0:i1))
      if(Debug > 0)Call VectorPrint("Flux",flux(:,k))
      enddo
   Call Report(xc,yc,wc,Ac,Bc,flux,errfx,errnm)
End Subroutine SolveBVP
! ----------------------------------------------------------------------
Function FluxCalc(rc,wc,Ac,Bc,Ax,rs,y,meth) result(f)
   Real(float) :: rc(0:),wc(0:),Ac(0:,0:),Bc(0:,0:),Ax(0:,0:),rs(0:),y(0:)
   Real(float) :: f(0:nflux),rx(0:size(y)-1)
   integer :: n,meth
   n =  size(y) - 2
   rx = rc*(1.0_float - y)
   f(0) = sum(wc*rx)       ! total flux, material balance
   f(4) = sum(Ac(0,:)*y)   ! first derivative, wrong for Lobatto
   f(5) = -sum(Ac(n+1,:)*y)
   f(6) = f(4) + f(5)
   if(meth == 1)then
   ! correction for conventional collocation
   ! no correction for Gauss, since endpoint wc = 0
      f(1) = f(4) + wc(0)*(rx(0) + sum(Bc(0,:)*y))
      f(2) = f(5) + wc(n+1)*(rx(n+1) + sum(Bc(n+1,:)*y))
   else
   ! calculation from equation coefficients (all but conventional)
      f(1) = rs(0) - sum(Ax(0,:)*y)
      f(2) = rs(n+1) - sum(Ax(n+1,:)*y)
   endif
   f(3) = f(1) + f(2)   ! this total equals f(0)
End Function FluxCalc
! ----------------------------------------------------------------------
Function FuncRX(x)  result(f)
   real(float), intent(in) :: x(:)
   real(float) :: f(size(x))
   ! calculate cubic variation of rate coefficient
   f = Aparm + Aparm1*(1.5_float - x)*x*x
End Function
!-------------------------------------------------------------------------------
Subroutine ReportBasic(xc,wc,rc,Ac,Bc,Cc,ityp)
   real(float) :: xc(:),wc(:),rc(:),Ac(:,:),Bc(:,:),Cc(:,:)
   integer :: n,ityp
   n = size(xc) - 2
   write(iout,'(/,3a,i3,a)') &
      "Solution using ",trim(TxtType(ityp))," points with",n," interior points"
   write(iout,'(a,2f10.4)') &
      'For Thiele Modulus and variation parameters of:',Thparm,Aparm/Thparm2
   if(.not.Dirichlet)write(iout,'(A,2F8.2)')'Left and Right Biot Nos:',Biot
   Call ArrayPrint('Quadrature points & weights, & Reactivity:',xc,wc,rc)
   if(Debug > 0)then
      Call ArrayPrint('A - 1st Derivative Operator',Ac)
      Call ArrayPrint('B - 2nd Derivative Operator',Bc)
      Call ArrayPrint('C - 2nd Derivative stiffness',Cc)
   endif
End Subroutine ReportBasic
! -----------------------------------------------------------------------
Subroutine Report(xc,y,wc,Ac,Bc,flux,errfx,errnm)
   real(float) :: xc(0:),y(0:,:),wc(0:),Ac(0:,0:),Bc(0:,0:)
   real(float) :: flux(0:,:),errfx(:,:),errnm(:,:)
   real(float) :: yex(0:size(xc)-1),ycout(0:size(xc)-1,2),rqc(0:size(xc)-1)
   Real(float) :: exflux(0:nflux),errflux(0:nflux,nsolv)
   real(float) :: errn(nsolv),errn1(nsolv,2),bcerr(0:1,nsolv),rerr(0:1,nsolv),ravg(2,nsolv)
   Real(float) :: yx(nx,nsolv),yxx(nx),rqx(nx),yout(nx,2)
   integer :: n,i,j,na
   integer, parameter :: nxtra=3
   Real(float) :: Lc(nx,size(xc)),fluxexact,an(0:size(xc)+nxtra,4)
   real(float) :: xx(nx)   ! temporary override use Xex values
   Character :: txtn*7
   Character (len=*), Parameter :: fmtv = '(25(f14.10,a1))'
   Character (len=*), Parameter :: fmtx = '(a,10(a1,f11.9))'
   Character (len=*), Parameter :: fmtg = '(a,10(a1,g12.5))'
   Character (len=9), parameter :: titlex(0:4)= &
      (/" Exact   ","Collocat.","Symmetric","Interp. r","Gal./Mom."/)
   Character (len=*), Parameter :: title = tb//"   x    "//tb//"     Exact"//tb// &
      "   Collocation"//tb//"   Symmetric"//tb//" Interpolated r"//tb//"Exact Mass"
   Character (len=*), Parameter :: titlef = tb//tb//"   Integrated"//tb//tb//" Left"//tb//tb// &
      "Right"//tb//tb//"Total"//tb//tb//" Left"//tb//tb//"Right"//tb//tb//"Total"
   ! -----------------------------------
   n = size(xc) - 2
   na = size(xc)+nxtra
   txtn =' n =xx"';  write(txtn(5:6),'(i2)')n
   flux = flux/Thparm2  ! normalize fluxes
   Ac(0,0) = Ac(0,0) - Biot2(0);
   Ac(n+1,n+1) = Ac(n+1,n+1) + Biot2(1);
   call ErrCalc(ConstCoef,Dirichlet,Thparm,Biot) ! tell error routine type of problem
   xx = Xex
   rqx = FuncRx(xx);  rqc = FuncRx(xc)
   bcerr = 0.0_float;  rerr = 0.0_float
   if(.not.Dirichlet)then     ! boundary condition residuals
      do j=1,nsolv
         bcerr(0,j) = sum(Ac(0,:)*y(:,j))
         bcerr(1,j) = -sum(Ac(n+1,:)*y(:,j))
      enddo
   endif
   ! ----- calculate interpolated solutions
   Lc = Lcoeff(xx,OC)
   yex = Exact(xc)               ! exact at xc & xx
   yxx = Exact(xx,exflux(0:2))   ! and exact flux
   fluxexact = sum(wc*(1.0_float - yex)*FuncRx(xc))/Thparm2
   exflux(3:6:3) = exflux(0); exflux(4:5) = exflux(1:2)
   do j=1,nsolv
      yx(:,j) = MatMul(Lc,y(:,j))
      errflux(:,j) = abs(flux(:,j)/exflux - 1.0_float)
   enddo
   errn = ErrorNorm(OC,y)
   errn1 = ErrorNorm1(OC,y)
   ! This output allows side to side comparison of 4 methods
   write(iout,'(/a)')"Results, exact and four methods: (exact if const. coef. or case in text)"
   Call ArrayPrint(title,xc,yex,y,fmtx=fmtv)
   write(iout,'(/a)')"Exact and Interpolated results:"
   Call ArrayPrint(title,xx,yxx,yx,fmtx=fmtv)
   ! Code for plotting with Tecplot remove comments to activate
   do j=1,nsolv   ! This output give more detail on each method
      yout(:,1) = yx(:,j) - yxx;  ycout(:,1) = y(:,j) - yex(:)             ! error y
      ycout(:,2) = MatMul(Bc,y(:,j));  yout(:,2) = MatMul(Lc,ycout(:,2))   ! 2nd Derivative
      yout(:,2) = yout(:,2) + rqx*(1.0_float - yx(:,j))  ! Residual
      ycout(:,2) = ycout(:,2) + rqc*(1.0_float - y(:,j)) ! Residual
      write(iout2,'(a,10(a,a1))')'# ',Titlex(j),tb,'x',tb,'y exact',tb,'yc',tb,'y err',tb,'Resid'
      !Call ArrayPrint('Zone T="'//Trim(TxtType(ityp))//txtn,xc,yex,y(:,j),ycout,fmtx=fmtv)
      Call ArrayPrint('Zone T="'//Trim(TxtType(ityp))//txtn,xx,yxx,yx(:,j),yout,flog=.false.,fmtx=fmtv)
      rerr(0,j) = yout(1,2)   ! residual at left boundary
      rerr(1,j) = yout(nx,2)  ! residual at right boundary
      an(:,j) = ErrTransform(na,xx,yout(:,2),ityp)
      ravg(:,j) = ErrResid(yout(:,2),ityp)
   enddo
   ! boundary condition residual and residual at boundary
   write(iout,'(/a)')'Boundary condition residual and residual at boundaries'
   do j=1,nsolv
      write(iout,'(a,i2,a1,i2,4(a1,g17.10))')'BCerr:' & ! bc res. x=0, R(0), bc res. x=1, R(1)
         ,n,tb,j,tb,abs(bcerr(0,j)),tb,abs(rerr(0,j)),tb,abs(bcerr(1,j)),tb,abs(rerr(1,j))
   enddo
   write(iout,'(/a)')"flux calculations, integrated, correct & 1st derivative:"
   write(iout,'(A)')titlef
   do j=1,nsolv
      write(iout,fmtx)titlex(j),(tb,flux(i,j),i=0,nflux)
   enddo
   write(iout,fmtx)titlex(0),(tb,exflux(i),i=0,nflux)
   write(iout,'(a)')"Flux Errors:"
   do j=1,nsolv
      write(iout,fmtg)titlex(j),(tb,errflux(i,j),i=0,nflux)
   enddo
   write(iout,'(a,f12.8,E15.7,a)')'flux:',fluxexact,abs(fluxexact/exflux(0)-1.0_float) , &
      ' quadrature of exact values'
   write(iout,'(/,a)')"Transforms of Residual:"
   write(iout,'(4(a1,2x,a,1x))')(tb,titlex(j),j=1,nsolv)
   do i=0,na
      write(iout,'(4(a1,g12.5))')(tb,an(i,j),j=1,nsolv)
   enddo
   call ArrayPrint('Residual norms, L2 & weighted:',ravg)
   Call VectorPrint('L2 error norms (Lobatto Quad.):',errn)
   Call VectorPrint('L2 error norms (Trap. Rule):',errn1(:,1))
   Call VectorPrint('L1 error norms (Trap. Rule):',errn1(:,2))
   errfx(1,:) = errflux(2,:)  ! 0 overall, 2 for right side, corrected
   errfx(2,:) = errflux(5,:)  ! right side, derivative
   errnm(1,:) = errn1(:,2)    ! L1 error norms (Trapezoidal Rule)
   errnm(2,:) = errn1(:,1)    ! L2 error norms (Trapezoidal Rule)
End Subroutine Report
!-------------------------------------------------------------------------------
Subroutine ReportErr(errx,nmin,nmax,ndel,title)
   real(float) :: errx(:,:)
   integer :: n,nmin,nmax,ndel,j
   Character (len=*) :: title
   Character (len=*), Parameter :: fmtg = '(i3,5(a1,g14.7))'
   write(iout,'(A)')title
   do n=nmin,nmax,ndel
      write(iout,fmtg)n,(tb,errx(j,n),j=1,nsolv)
   enddo
End Subroutine ReportErr
!-------------------------------------------------------------------------------
End Module BVPmod
!-------------------------------------------------------------------------------
Program Example01  ! Nonsymmetric BVP - particle problem
   Use OrthogonalColloc, Only : Gauss,Lobatto,Chebyshev,RadauR,RadauL
   Use BVPmod, Only : SolveBVP, ReportErr, nsolv
   Use Array_Print, Only : OpenFile
   include 'defs.fi'
   Integer :: ityp,n,Nmin,Nmax, Ndel   ! type, no., min, max, delta coll. points
   Character :: fname*(80),ch*1
   integer, parameter :: MaxNmax = 50
   real(float) :: Thparm,Aparm,BCflux(2)
   real(float) :: errnorm(2,nsolv,MaxNmax)
   real(float) :: errflux(2,nsolv,MaxNmax)
   logical :: variable
   Character (len=*), Parameter :: fmtf ='(i3,12(a1,f16.12))'
   ! ----------------------------------------------------------------
   ityp = Gauss;  Nmin = 2;   Nmax = 2;   Ndel = 1
   ThParm = 10.0_float;   AParm = 1.0_float; BCflux = 0.0_float
   write(*,'(a)')' Enter Thiele Modulus and Biot no.>'
   read(*,*)Thparm, BCflux(1)
   write(*,'(a)')' Variable coefficient (T/F)? >'
   read(*,'(L5)')variable
   if(variable)Aparm = 0.2_float
   BCflux(2) = BCflux(1)  ! could be different
   write(*,'(a,i2,a,i2,a)')' Enter g/l/c/r/x for Gauss/Lobatto/Chebyshev/Radau(R/L) >'
   read(*,*)ch;  ityp = index('glcrx',ch);  ityp = Max(Gauss,Min(RadauL,ityp))
   write(*,'(a)')' Enter min, max & delta N>'
   read(*,*)Nmin,Nmax,Ndel
   Nmin = max(Nmin,1)
   Nmax = min(max(Nmin,Nmax),MaxNmax)
   write(*,'(A)')' Enter output file name>'
   read(*,*)fname
   Call OpenFile(fname)
   Call OpenFile(fname,append='2')
   ! ---------------------------------------------------
   do n=Nmin,Nmax,Ndel
      Call SolveBVP(n,ityp,Thparm,Aparm,BCflux,errflux(:,:,n),errnorm(:,:,n))
   enddo
   Call ReportErr(errflux(1,:,:),Nmin,Nmax,Ndel,'Corrected Flux Error')
   Call ReportErr(errflux(2,:,:),Nmin,Nmax,Ndel,'Derivative Flux Error')
   Call ReportErr(errnorm(1,:,:),Nmin,Nmax,Ndel,'L1 Error Norms')
   Call ReportErr(errnorm(2,:,:),Nmin,Nmax,Ndel,'L2 Error Norms')
End Program Example01

