Program Jacobi_test
   Use Orth_Poly, only : x_jacobi,d_jacobi,Jacobi_All,Jacobi_Deriv,Jacobi_Deriv1,Jacobi_Deriv2, &
                        Jacobi_Ends,RootEstimate
   Use Array_Print, only : ArrayPrint, VectorPrint, OpenFile
   include 'defs.fi'
   integer :: n,nx
   real(float) :: a,b
   logical :: monc,shft,ultra
   Integer, Parameter :: N01 = 28   ! used by Xcalc
   Real(float), Parameter :: pi = 3.14159265358979323_float
! -----------------------------------------------------------------
!  Testing of some fundamental calculations for better understanding
!  All of these calculations are described in Chapter 2
!  JacobiTest does the following:
!     - calculation recursion and other fundamental coefficients
!     - coefficients of derivatives of Jacobi polynomials
!     - tabular values of Jacobi polynomials & derivatives 1 thru 4
!     - roots of Jacobi polynomials by 2 methods
!     - prints a table of results
!  Results are in a file polytest.dat
! -----------------------------------------------------------------
   Call OpenFile('polytest')
   a = 0.0_float;  b = 0.0_float
   n = 1;  monc=.false.;  shft=.false.
   write(*,*)'Enter n,a,b >'
   read(*,*)n,a,b
   nx = n
   ultra = a==b
   if(ultra)nx = n/2
   monc=.false.;  shft=.false.
   Call JacobiTest(n,nx,a,b)        ! test various routines in Orth_Poly
   monc=.false.;  shft=.true.
   Call JacobiTest(n,nx,a,b)        ! test various routines in Orth_Poly
   monc=.true.;  shft=.false.
   Call JacobiTest(n,nx,a,b)        ! test various routines in Orth_Poly
   monc=.true.;  shft=.true.
   Call JacobiTest(n,nx,a,b)        ! test various routines in Orth_Poly
!----------------------------------------------------------
   contains
!----------------------------------------------------------
   Subroutine JacobiTest(n,nx,a,b)
      Use Orth_Poly
      Use Array_Print, only : ArrayPrint, VectorPrint
      include 'defs.fi'
      integer :: n,nx,nt
      Real(float) :: a,b,c1(3),c2(3),c3(3),c4(3),P0(0:1),fshft
      Real(float) :: xr(nx,0:2),ax(0:n,5),dj(0:n-1,0:n-1)
      Real(float) :: x(N01+nx),x2(N01+nx),Pn(N01+nx,0:n),DPn(N01+nx,0:4)
      Character (len=*), Parameter :: fmts = '(25(g14.7,a1))' !  '(25(es24.17,a1))'
      !-------------------------------------------------------------------------
      nt = N01+nx
      write(*,'(/a,2f4.1,a,i4,a,3L2)')'**** Orth_Poly Test, a,b =', &
          a,b,', n =',n,' Monic, Shift, Ultra =',monc,shft,ultra
      write(iout,'(/a,2f4.1,a,i4,a,3L2)')'**** Orth_Poly Test, a,b =', &
          a,b,', n =',n,' Monic, Shift, Ultra =',monc,shft,ultra
      !----------------- Calculation of recurrence coefficients ----------------
      ax(:,1:4) = x_jacobi(n+1,(/a,b/),Monic=monc,Shift=shft)   ! recursion coefficients
      ax(:,5)= d_jacobi(n,(/a,b/),Shift=shft)                   ! integral Pn^2
      call ArrayPrint('Recursion & Other Coefficients:',transpose(ax))
      ! ------------------ Calculation of Roots --------------------------------
      xr(:,1) = JacobiRootEstimate(n,nx,a,b) ! simple estimate
      xr(:,0) = RootEstimate(n,nx,(/a,b/))       ! more accurate
      xr(:,0) = cos(xr(:,0))
      if(abs(a)==0.5_float .and. abs(b) == 0.5_float)then
         xr(:,1) = xr(:,0)
         xr(:,2) = xr(:,0)
      else
         xr(:,1) = Jacobi_Nroots(n,a,b,xr(:,0))    ! Newton-Raphson w/root suppresion
         xr(:,2) = Jacobi_Eroots(n,nx,a,b)         ! Eigenvalues
      endif
      call ArrayPrint('Roots estimates, 2 solutions',transpose(xr),fmtx='(25(f20.16,a1))')
      ! ----------- Calculation of Polynomials & Derivative --------------------
      x = Xcalc(xr(:,1))  ! calculates values for tabulation includes roots
      if(shft)then
         fshft = 2.0_float
         x = (x + 1.0_float)*0.5_float
         x2 = max(x*(1.0_float - x),Epsilon(x))
      else
         fshft = 1.0_float
         x2 = max(1.0_float - x**2,Epsilon(x))
      endif
      Pn = Jacobi_All(x,n,arec=ax(:,1:4),Monic=monc)
      dj = Jacobi_Deriv(n,(/a,b/),ax(:,1:4),Monic=monc,Shift=shft)   ! coefficients for derivatives P1 thru Pn
      DPn(:,0) = MatMul(Pn(:,0:n-1),dj(n-1,:))    ! 1st derivative of Pn(n)
      Call ArrayPrint('Jacobi 1st Derivative Coefficients:',dj)
      P0 = Jacobi_Ends(n-1,(/a+1.0_float,b+1.0_float/),Monic=.false.,Shift=shft)
      P0 = fshft*(0.5_float)*(Real(n+1,float)+a+b)*P0
      if(monc)P0 = P0/ax(n,3)  ! divide by lead coef., easiest fix
      c2 = Jacobi_Deriv2(n,(/a,b/),Shift=shft)
      c3 = Jacobi_Deriv2(n,(/a,b/),nd=3,Shift=shft)
      c4 = Jacobi_Deriv2(n,(/a,b/),nd=4,Shift=shft)
      c1 = Jacobi_Deriv1(n,(/a,b/),Monic=monc,Shift=shft)
      write(iout,'(a,4(a,g14.7))')'Derivative ends:',tb,P0(0),tb,P0(1)
      write(iout,'(a,4(a,g14.7))')'Jacobi_Deriv1  :',tb,c1(1),tb,c1(2),tb,c1(3)
      write(iout,'(a,4(a,g14.7))')'Jacobi_Deriv2  :',tb,c2(1),tb,c2(2),tb,c2(3)
      write(iout,'(a,4(a,g14.7))')'Jacobi_Deriv3  :',tb,c3(1),tb,c3(2),tb,c3(3)
      write(iout,'(a,4(a,g14.7))')'Jacobi_Deriv4  :',tb,c4(1),tb,c4(2),tb,c4(3)
      DPn(:,1) = ((c1(1)*x + c1(2))*Pn(:,n) + c1(3)*Pn(:,n-1))/x2 ! 1st derivative
      DPn(:,2) = ((c2(1)*x + c2(2))*DPn(:,1) + c2(3)*Pn(:,n) )/x2 ! 2nd
      DPn(:,3) = ((c3(1)*x + c3(2))*DPn(:,2) + c3(3)*DPn(:,1))/x2 ! 3rd
      DPn(:,4) = ((c4(1)*x + c4(2))*DPn(:,3) + c4(3)*DPn(:,2))/x2 ! 4th
      Dpn(1,1) = P0(0)
      Dpn(nt,1) = P0(1)
      call ArrayPrint("Polynomial Values & Derivative:",x,Pn,DPn,fmtx=fmts)
   End Subroutine JacobiTest
   !----------------------------------------------------------------------------
   Function Jacobi_Nroots(n,a,b,x0)  result(xx)
      integer, intent(in) :: n
      Real(float), intent(in) :: a,b
      real(float) :: x0(0:),xx(0:size(x0)-1)
      Real(float) :: dx,x,dp,dq,ab(0:n-1,4)
      integer :: nx,i,ntot
      real(float), parameter :: xtol = Epsilon(dx)*100.0_float
      ! --------------------------------
      ! This program calculates Jacobi roots using Newton-Raphson iteration with
      ! suppression of roots found.
      !  x_Jacobi gives the recurrence coefficients, like the matlab OPQ code
      !  JacobiRootEstimate - estimates the roots
      !  Pcalc - calculated p/p' where p is the polynomial
      ! The estimation routine is so accurate that suppression term
      ! contributes little to the iteration, so dx ~ dp
      ! --------------------------------
      nx = size(x0)
      ab = x_jacobi(n,(/a,b/),Monic=.true.)       ! get the recurrence coefficients
      ntot = 0
      do i=0,nx-1
         x = x0(i)   ! start with estimate  dq = Qcalc(x,xx(0:i-1))
         do          ! newton iteration
            dp = Pcalc(x,ab);  dq = sum(1.0_float/(x - xx(0:i-1)))
            dx = dp  !(1.0_float - dp*dq)
            x = x - dx; ntot = ntot + 1
            if(abs(dx) < xtol)exit
         enddo
         xx(i) = x
      enddo
      write(iout,'(a,i4)')'Total interations =',ntot
   End Function Jacobi_Nroots
   ! ----------------------------
   Function Pcalc(x,ab) result(dp)
      Real(float) :: x,dp,ab(0:,:)
      Real(float) :: p2,p0,p1,pp2,pp1,pp0
      integer :: k,n
      ! calculate Jacobi polynomial and derivatives for Jacobi_Nroots
      n = size(ab,1)
      p1 = 1.0_float;   p2 = x - ab(0,1)
      pp1 = 0.0_float;  pp2 = 1.0_float
      do k = 1,n-1
         p0 = p1;  p1 = p2;  pp0 = pp1;  pp1 = pp2
         p2 = (x - ab(k,1))*p1 - ab(k,2)*p0
         pp2 = p1 + (x - ab(k,1))*pp1 - ab(k,2)*pp0
      enddo
      !write(iout,'(a,10g14.7)')'x,pn,pn-1,pn-2:',x,p2,p1,p0
      dp = p2/pp2
   End Function Pcalc
   !----------------------------------------------------------------------------
   Function Jacobi_Eroots(n,nx,a,b)  result(xr)
      Use EigenValue, Only : EigenValTriSym
      integer, intent(in) :: n,nx
      Real(float), intent(in) :: a,b
      Real(float) :: x(0:n-1),ab(0:n-1,4),xr(0:nx-1)
      ! --------------------------------
      ! This program calculates Jacobi roots using the eigenvalues technique.
      ! it uses:
      !   r_Jacobi which mimics the matlab OPQ code
      !   EigenValTriSym is a wrapper around DSTEQR an LAPack routine in EigenLA
      ! DSTEQR updates the Eispack routine used by Elkay and Kautsky
      ! Could probably do better with an iterative method.
      ! --------------------------------
      ab = x_Jacobi(n,(/a,b/),Monic=.true.)
      ab(:,2) = sqrt(ab(:,2))
      x = EigenValTriSym(ab(:,1),ab(1:,2))      ! calls DSTEQR
      xr = x(n-nx:n-1)
   End Function Jacobi_Eroots
   !----------------------------------------------------------------------------
   Function JacobiRootEstimate(n,nx,a,b)   result(x)
      Integer, intent(in) :: n,nx
      Real(float), intent(in) :: a,b
      Real(float) :: x(nx),zn(nx),abn
      Integer :: k
      ! -------------------------------
      ! Simple root estimates using generalized Chebyshev formula
      ! -------------------------------
      zn = (/(Real(2*k,float),k=n-nx+1,n)/)
      abn = Real(2*n+1,float) + a + b
      zn = pi*(zn + b - 0.5_float)/abn
      x = -cos(zn)
   End Function JacobiRootEstimate
   !----------------------------------------------------------
   Function Xcalc(x1) result(x)
      real(float) :: x1(:),x(size(x1)+N01)
      integer :: n1,n2,i1,i2,k
      ! for following requires N01 = 28
      Real(float), Parameter :: x2(N01) = (/-1.0D0,-0.99D0,-0.97D0,-0.94D0,-0.90D0, &
         (real((/((k),k=0,17)/),float)-8.5D0)*0.1D0,0.996D0,0.997D0,0.998D0,0.999D0,1.0D0/)
      !x2 = -cos(pi*real((/((k),k=0,N01-1)/),float)/real(N01-1))  ! works for any N01
      i1 = 1;  n1 = size(x1)
      i2 = 1;  n2 = size(x2)
      do k=1,n1+n2
         if(x1(i1) < x2(i2))then
            x(k) = x1(i1)
            i1 = i1 + 1
            if(i1 > n1)exit
         else
            x(k) = x2(i2)
            i2 = i2 + 1
            if(i2 > n2)exit
         endif
      enddo
      if(k < n1+n2)then
         if(i1 > n1)x(k+1:) = x2(i2:)
         if(i2 > n2)x(k+1:) = x1(i1:)
      endif
   End Function Xcalc
   !----------------------------------------------------------
End Program Jacobi_test

