Program Jacobi_test
   Use Orth_Poly, only : Legendre_Transform, Jacobi_Transform, Jacobi_Series, Polyset, Jacobi_Scale_Sym
   Use Orth_Cheb, only : Cheb_Scale
   Use OrthogonalColloc, Only : ColDataType, Initialize, Xpoints, WeightQ, Lcoeff, &
                        Gauss, Lobatto, Chebyshev, RadauR, RadauL, Chebyshev1, TxtType, &
                        Nonsymmetric, Planar, Cylindrical, Spherical
   Use Array_Print, only : ArrayPrint, VectorPrint, OpenFile
   include 'defs.fi'
   Type(ColDataType) :: OC
   integer :: n,FuncCase
   logical :: shft
   integer :: geom
   integer, parameter :: Nfunc = 6
   Character (len=30) :: FuncTxt(Nfunc)
   Character (len=*), Parameter :: fmt12 = '(25(f12.8,a1))'
! -----------------------------------------------------------------
!  Testing of some fundamental calculations for better understanding
!  All of these calculations are described in Chapter 2, some use data from
!  Chapter 3 examples.
!  TranFunc:
!     calculates transform coefficients for Nfunc = 6 different functions and
!     then uses the transform in the Jacobi series to check the results
!  TransformTest:
!     calculates Legendre series coefficients from 4 sets of nodal values
!     2 cases are at Gauss points, 2 are at Lobatto points
!     The values are checked by recalculing the nodal values
!     using the Legendre series.
! -----------------------------------------------------------------
   Call OpenFile('polytran') ! store results in polytran.dat
   geom = 0
   write(*,'(a,4I2,a)')' Enter',Nonsymmetric,Planar,Cylindrical,Spherical,&
      ' for Nonsymmetric, Planar, Cylindrical, or Spherical geometry>'
   read(*,*)geom
   shft = .false. ! manually set or add to read
   if(geom > 0)shft = .true.
   write(*,*)'Enter n >'
   read(*,*)n
   Call TranFunc(n)     ! calculate transforms coefficients of various functions
   Call TransformTest   ! calculate Legendre coefficients from nodal values
!----------------------------------------------------------
   contains
!----------------------------------------------------------
   Subroutine TransformTest
      integer, parameter :: n = 4
      real(float),parameter :: ymom(0:5) = (/0.0,0.3282055029,0.9372069920,0.9437745518,0.6326982777,0.0/)
      real(float),parameter :: yglk(0:5) = (/0.0,0.4660299556,0.9273132387,0.9708737656,0.7906926421,0.0/)
      real(float),parameter :: ylob(0:5) = (/0.33498363,0.79089770,0.98104193,0.98104193,0.79089770,0.33498363/)
      real(float),parameter :: ygau(0:5) = (/0.32360985,0.67518982,0.97476460,0.97476460,0.67518982,0.32360985/)
      Real(float),parameter :: yxx(0:5,2,2) = Reshape((/ymom,ygau,yglk,ylob/),(/6,2,2/))
      real(float) :: Q(0:n+1,0:n+1),at(0:n+1,0:2),x(0:n+1),w(0:n+1)
      integer :: k,ityp
      ! ---------------------------------
      ! This routine tests Legendre tranforms using Gauss & Lobatto points
      ! It uses the 4 sets of nodal values above.
      ! Each set is used to calculate Legendre coefficients, then the
      ! nodal values are recalculated from the series for comparison.
      ! solutions are from Ch 3 for catalyst particle problem with n = 4
      ! values at Gauss points:
      !    ymom - Method of moments for variable coefficient problem, fig. 3.15
      !    ygau - Bi = 10, constant coefficient problem Gauss (same as moments), fig. 3.9
      ! values at Lobatto points:
      !    yglk - Galerkin method for variable coefficient problem, fig. 3.16
      !    ylob - Bi = 10, constant coefficient problem Lobatto (same as Galerkin), fig. 3.9
      ! after the Legendre polynomial coefficients are calculated, the results are checked
      ! by calculating the nodal values with the Legendre series
      ! Caution: beware of Shift defaults OrthogonalColloc -> .true, Orth_Poly -> .false.
      ! ---------------------------------
      shft = .true.  ! either will work (see caution above)
      ! ------------ Test 1 Lobatto points ---------------------
      write(iout,'(/a,2f6.2,a,i3,2(a,L2))')'**** Test of Transforms, Gauss pts, then Lobatto pts.'
      do ityp = Gauss,Lobatto
         call Initialize(OC,n,TType=ityp,Shift=shft)
         x = Xpoints(OC)
         w = WeightQ(OC)
         Q = Legendre_Transform(x,w,ityp,Shift=shft)
         Call ArrayPrint('Legendre transform matrix '//Trim(TxtType(ityp))//' points:',Q,fmtx=fmt12)
         do k = 1,2
            at(:,0) = MatMul(Q,yxx(:,k,ityp))  ! Legendre coeff.
            at(:,1) = yxx(:,k,ityp)            ! nodal values
            at(:,2) = Jacobi_Series(at(:,0),x,Shift=shft)  ! Legendre series
            Call ArrayPrint('Pn Coefficient, Original y, Reproduced y:',Transpose(at),fmtx=fmt12)
         enddo
      enddo
   End Subroutine TransformTest
   ! -------------------------------------------
   Subroutine TranFunc(n)
      integer, intent(in) :: n
      integer :: k,t
      integer, parameter :: t0=Gauss,t1=Chebyshev1,nval = 10
      real(float), parameter :: xcalc(0:nval) = Real((/(k,k=0,nval)/),float)/Real(nval,float)
      real(float) :: xc(0:nval),xf(0:nval),f(0:nval),fp(0:nval,t0:t1)
      real(float) :: ab(2),a1(0:n,t0:t1),af(0:n,t0:t1)
      ! --------------------------------------------------
      !  This routine calculates Jacobi transforms for several functions
      !  Others could easily be added to the FuncTran function.
      !  The code calculates the Jacobi transforms for all the types supported
      !  The code then evaluates the Jacobi series and compares the results to
      !  the actual function.
      !  Polyset - gives alpha and beta for the type and geometry
      !  Cheb_Scale - gives factors to convert between Chebyshev & Jacobi scaling
      !  Jacobi_Scale_Sym - give factors to convert between symmetric Jacobi polynomials
      !                     and equivalent nonsymmetric ones
      !  Jacobi_Transform - evaluates the transform coefficients
      !  Jacobi_Series  - uses the tranform coefficients to calculate function values
      ! --------------------------------------------------
      write(iout,'(/a,i3,a,L2)')'**** Test Jacobi_Transform & Cheb_Transform, n =',n,', Shifted =',shft
      write(iout,'(/a)')'Transforms considered and (alpha,beta) values:'
      a1 = 0.0_float;    fp = 0.0_float;  af = 1.0_float
      xc = xcalc;        xf = xcalc
      if(.not.shft)then
         xc = xcalc*2.0_float - 1.0_float
         xf = xc
      elseif(geom > 0)then
         xc = xcalc**2
      endif
      do t = t0,t1
         ab = Polyset(t,geom)
         if(geom > 0 .and. (t==RadauR .or. t==RadauL))cycle
         write(iout,'(a1,a,a1,2f5.1)')tb,TxtType(t),tb,ab
         if(geom .ne. Nonsymmetric)then
            af(:,t) = 1.0_float/Jacobi_Scale_Sym(n,ab)
         endif
      enddo
      a1(:,t0) = 1.0_float/Cheb_Scale(n,Polyset(Chebyshev,geom))
      a1(:,t0+1) = 1.0_float/Cheb_Scale(n,Polyset(Chebyshev1,geom))
      call arrayprint('Chebyshev/Jacobi Scaling factors - 2nd, 1st:',a1(:,t0:t0+1),fmtx=fmt12)
      if(geom > 0) &
         call arrayprint('Symmetric coefficient scaling factors',af(:,t0:t1),fmtx=fmt12)
      do FuncCase=1,Nfunc   ! loop over functions
         do t = t0,t1      ! point types
            if(geom > 0 .and. (t==RadauR .or. t==RadauL))cycle
            ab = Polyset(t,geom)  ! alpha & beta
            a1(:,t) = Jacobi_Transform(n,t,geom,FuncTran,shft)          ! calculate transform
            fp(:,t) = Jacobi_Series(a1(:,t),xc,abeta=ab,Shift=shft)   ! evaluate Jacobi series
         enddo
         where(abs(a1) < 1.0e-13_float)a1 = 0.0_float
         write(iout,'(/5a)')'Coefficients for f(x) = ',FuncTxt(FuncCase)
         write(iout,'(10(a12,a1))')(trim(TxtType(t)),tb,t=t0,t1)
         call arrayprint(' ',a1(:,t0:t1),fmtx=fmt12)
         f = FuncTran(xf)
         Call arrayprint('Tabular Value Check, x, f(x), Jacobi Series',xcalc,f,fp(:,t0:t1),fmtx=fmt12)
      enddo
   End Subroutine TranFunc
   ! -------------------------------------------
   Function FuncTran(xx) result(f)
      include 'defs.fi'
      Real(float), intent(in) :: xx(:)
      Real(float) :: f(size(xx))
      integer :: nx,i
      Real(float) :: x(size(xx)),p(size(xx),0:4),t(size(xx)),pfac(0:4)
      Real(float), parameter :: a0 = 0.20_float, a1 = (1.0_float - a0)*4.0_float
      ! ---------------------------------------------------
      !  Functions for testing Jacobi transforms
      !  Note: pfac is used to scale between conventional Chebyshev - Jacobi
      !  using scaling factors calculated by Cheb_Scale
      ! ---------------------------------------------------
      nx = size(x)
      x = xx
      if(shft .and. geom == 0)x = 2.0_float*x - 1.0_float
      Select case(FuncCase)
      case(1)  ! example in Table 2.6, section 2.8
         FuncTxt(1) = '1 - 0.6x^2 - 0.4x^4'
         f = (1.0_float - x*x)*(1.0_float + x*x*0.4_float)
      case(2)  ! first 4 Jacobi(1,1) polynomials
         FuncTxt(2) = 'P0 + 0.4P2 + 0.2*P4 (Lobatto)'
         p(:,0) = 1.0_float
         p(:,1) = x*2.0_float
         p(:,2) = 1.875_float*x*p(:,1) - 0.75_float*p(:,0)
         p(:,3) = 1.86666666666667_float*x*p(:,2) - 0.8_float*p(:,1)
         p(:,4) =  1.875_float*x*p(:,3) - 0.833333333333333_float*p(:,2)
         f = p(:,4)*0.2_float + p(:,2)*0.40_float + p(:,0)
      case(3)  ! exponential function
         FuncTxt(3) = 'exp(-5x^2)'
         f = exp(-5.0_float*x*x)
      case(4)  ! Chebyshev 2nd kind, Jacobi scaling
         FuncTxt(4) = 'P0 - .4P2 + .05P4 (Cheby 2nd)'
         t = acos(x)
         pfac = 1.0_float/(/1.0_float,1.33333333_float,1.6_float,1.82857143_float,2.03174603_float/)
         p(:,0) = 1.0_float
         do i = 1,4
            p(:,i) =(sin(real(i+1,float)*t)/sin(t))
            if(x(nx) > 0.99999999_float)p(nx,i) = real(i+1,float)
         enddo
         forall(i=0:4)p(:,i) = p(:,i)*pfac(i)
         f = p(:,0) - p(:,2)*0.4_float + p(:,4)*0.05_float
      case(5)  ! exponential function
         FuncTxt(5) = 'Residual function'
         f = ResidFunc(x)
      case default  ! cubic coefficient of bvp in Ch. 3, varies 0.2 to 1.8
         if(geom == 0)then
            ! nonsymmetric function used in catalyst pellet problem with variable coeff.
            ! see section "Orthogonal Polynomial - Modal Trial Functions"
            FuncTxt(6) = '0.2 + 1.6x^2(3 - 2x)' ! nonsymmetric function
            x = xx;  if(.not.shft)x = (xx+1.0_float)*0.5_float
            f = a0 + a1*(1.5_float - x)*x*x
         else  ! use a symmetric function for this case
            FuncTxt(6) = 'T0 + 0.3T2 - .05T4 (Cheby 1st)'
            ! Jacobi(-0.5,-0.5) must use scaling factors from Jacobi_Scale_Sym,
            pfac = 1.0_float/(/1.0_float,2.0_float,2.66666667_float,3.2_float,3.65714286_float/)
            p(:,0) = 1.0_float   ! these have conventional Chebyshev scaling
            p(:,1) = x
            p(:,2) = (2.0_float)*x*p(:,1) - p(:,0)
            p(:,3) = (2.0_float)*x*p(:,2) - p(:,1)
            p(:,4) = (2.0_float)*x*p(:,3) - p(:,2)
            forall(i=0:4)p(:,i) = p(:,i)*pfac(i)
            f = p(:,0) + 0.3_float*p(:,2) - 0.05_float*p(:,4)
         endif
      end select
   End Function FuncTran
  !----------------------------------------------------------
   Function ResidFunc(x) result(R)
      include 'defs.fi'
      Real(float), intent(in) :: x(:)
      Real(float) :: R(size(x))
      Integer, parameter :: nx=18,trsd=1 ! < set this to 1 Moment, 2 Galerkin
      real(float) :: Li(size(x),0:nx+1)
      ! --------------------------------------------------------
      ! These are residuals from the variable coefficient problem of chapter 3 (code Ex01.f90)
      ! The residuals are the values at n = 18 Lobatto points
      ! The results are Moments & Galerkin methods for n = 6
      ! The residual is orthogonal to various lower order polynomials.
      ! The residual is a 10th order polynomial for this problem
      ! These results are discussed in section 3.1.6. Fig 3.40 shows the Galerkin results
      ! --------------------------------------------------------
      Real(float), parameter :: RsdM(0:nx+1) = &    ! moments, n=6, exact, i.e. +2 quadrature points
         (/ 7.2829563448698291d0, 4.8637124592352121d0, 0.7165637355934664d0,-2.5038021327672162d0,-2.9727347785418541d0, &
           -0.8747256839791440d0, 1.8577502396446184d0, 2.9164260874962786d0, 1.2392865340372297d0,-1.9225447700950260d0, &
           -3.5361256711865829d0,-1.4579128802228369d0, 2.7462115928952504d0, 4.5401539986576980d0, 1.1958016716573499d0, &
           -4.4171626009912917d0,-6.0179725695385198d0,-0.2949533358380623d0, 9.3384425255554788d0,15.4510277490779675d0  /)
      Real(float), parameter :: RsdG(0:nx+1) = &    ! Galerkin, n=6, inexact +1 quadrature points
         (/13.2817176692171124D0,10.3684639100667493D0, 5.0954857379284455D0, 0.2619831053043296D0,-2.0129368344025842D0, &
           -1.4996996793397059D0, 0.3106635460696818D0, 1.4501827446591760D0, 0.8646931621924114D0,-0.7914673268826906D0, &
           -1.6036129457354125D0,-0.3127716138768093D0, 1.8195479938638845D0, 1.8461693047243664D0,-1.2309912676332679D0, &
           -4.0544112032182866D0,-1.3875514056201581D0, 8.4476567529128772D0,20.9074883940394898D0,28.2136422507586246D0 /)
      Real(float),parameter :: Rsd(0:nx+1,2) = Reshape((/RsdM,RsdG/),(/nx+2,2/))
      ! --------------------------------------------------------
      call Initialize(OC,nx,TType=Lobatto,Geometry=Nonsymmetric,Shift=.false.)
      Li = Lcoeff(x,OC)
      R = Matmul(Li,Rsd(:,trsd))
   End Function ResidFunc
   !----------------------------------------------------------
End Program Jacobi_test
     

