Module TestModule
   Use OrthogonalColloc, Only : ColDataType,Initialize,Xpoints,WeightQ,WeightB,Deriv1,   &
         Deriv2,Stiffness,MassMatrix,DerivOdd,Lcoeff,Lpoly,DataDelete,pi, &
         Nonsymmetric, Planar, Gauss, Lobatto, Chebyshev, RadauR, RadauL, General
   Use Array_Print, Only : ArrayPrint, VectorPrint
   include 'defs.fi'
   Private
   public  TestData
   logical, public :: Symmetric,shifted,testm,testi
   integer, public :: igeom
   integer, parameter :: ni = 21 ! interpolation points for output
   Type(ColDataType) :: OC
   Character :: TxtFunc*(18)
   Contains
   ! ------------------------------------------------
   Subroutine TestData(n,typ,sym)
      Integer, intent(in) :: n,typ,sym
      Real(float) :: x(sym:n+1),w(sym:n+1),wb(sym:n+1),fwb
      Real(float) :: A(sym:n+1,sym:n+1),B(sym:n+1,sym:n+1),C(sym:n+1,sym:n+1),An(1:n+1,1:n+1)
      Real(float) :: D(sym:n+1,sym:n+1),Dx(sym:n+1,sym:n+1),Di(sym:n+1,sym:n+1)
      Real(float) :: q(sym:n+1,sym:n+1),fx(sym:n+1)
      Real(float) :: xi(ni),Li(ni,sym:n+1),Lp(ni,sym:n+1)
      Real(float) :: fint(3),fi(ni,3),dfi(ni,1),f(sym:n+1),df(sym:n+1,3)
      Integer :: k,which,nchk
      ! formats --------------------------------------------------------
      Character (len=*), Parameter :: fmf17 = '(25(f22.17,a1))'
      Character (len=*), Parameter :: fme17 = '(25(es24.17,a1))'
      !-------  Get the collocation data ------------------------------------------
      Symmetric = igeom >= Planar
      if(typ == General)then  ! equal spaced user supplied points
         x = Real((/(k,k=0,n+1)/),float)/Real(n+1,float)
         if(.not.shifted)x = x*2.0_float - 1.0_float
         call Initialize(OC,x,Geometry=igeom,Shift=shifted)
      else
         call Initialize(OC,n,TType=typ,Geometry=igeom,Shift=shifted)
      endif
      x  = Xpoints(OC)           ! collocation points
      w  = WeightQ(OC)           ! quadrature weights (new code)
      wb = WeightB(OC,fwb)       ! barycentric weights (new code)
      A = Deriv1(OC)             ! 1st derivative operator
      if(Symmetric)An = DerivOdd(OC)
      B = Deriv2(OC)             ! 2nd derivative (Laplacian) operator
      C = Stiffness(OC)          ! stiffness matrix
      D = MassMatrix(OC)         ! mass matrix no function, f(x)=1
      if(testm)then     ! test mass matrix calculations
         fx = MassFunc(x)
         Di = MassMatrix(OC,fx)            ! mass matrix, interpolated coefficient
         if(Symmetric)then
            ! MassFunc = 1 - x^2 for symmetric problem
            Dx = MassMatrix(OC,MassFunc)   ! mass matrix, function supplied, 1 extra pt.
         else
            ! MassFunc = 0.2 + 1.6(3 - 2x)x^2
            Dx = MassMatrix(OC,MassFunc,2) ! mass matrix, function supplied, 2 extra pts.
         endif
      endif
      nchk = QuadCheck(x,w,sym)
      if(testi)then
         which = 1   ! manually set to which function you want
         ! which = 1 exp(-5x^2), function ExpFunc
         ! which = 2 cosh(5x)/cosh(5, function Ex01Func
         ! which = 3 1/(1 + 25x^2), function RungeFunc
         Call TestInterpolation(which,x,w,A,xi,Li,q,Lp,f,df,fi,dfi,fint,sym)
      endif
      ! -------- Write to file (tab delimited) ------------------------------------
      Call Result_Print
      call DataDelete(OC)
   ! ------------------------------------------------------------------------------
   Contains ! the following:
   !  Result_Print - tabular results
   ! next 3 are interchangeable for testing integration, differentiation & interpolation
   !  ExpFunc - exp(-ax^2), integrals, values, derivatives for testing
   !  Ex01Func - cosh(ax)/cosh(a)  catalyst pellet problem
   !  RungeFunc - 1/(1 + ax^2) the Runge function
   !  MassFunc - function to demonstrate MassMatrix call
   !  EvalPoly - evaluates monomial form of interpolating polynomial
   !  QuadCheck - checks accuracy of quadrature
   ! ------------------------------------------------------------------------------
      Subroutine Result_Print
         Real(float) :: xlog
         Character*(15) :: cfac
         Character (len=*), Parameter :: fmti = '(a,a1,i3,a1,F10.5,2(a1,F14.8),a1,ES14.7)'
         Character (len=*), Parameter :: fmta = '(20(f12.8,a1))'
         xlog = Real(n,float)*log10(Real(n,float))
         write(cfac,'(3g15.8)')fwb
         Call Arrayprint('Points, Quad. Wts., Bar. Wts, factors ='//cfac,x,w,wb,fmtx=fmf17)
         write(iout,'(a,i3)')'Integration exact thru degree',nchk
         Call ArrayPrint('A - Derivative:',A)
         if(Symmetric)Call ArrayPrint('An Odd Matrix:',An)
         Call ArrayPrint('B - Laplacian:',B)
         Call ArrayPrint('C - Stiffness Matrix:',C)
         Call ArrayPrint('D - Mass Matrix - f(x) = 1:',D)
         if(testm)then
            call vectorprint('MassFunc',fx)
            Call ArrayPrint('Di - Mass Matrix Interpolated:',Di)
            Call ArrayPrint('Dx - Mass Matrix - f(x):',Dx)
         endif
         if(testi)then
            call ArrayPrint('Lagrange Interpolating Polynomials:',xi,Li,fmtx=fmta)
            ! remove comment to print
            !call ArrayPrint('Interpolating Polynomials from Monomials:',xi,Lp,fmtx=fmta)
            !call ArrayPrint('Interpolating Monomials:',q)
            write(iout,'(2a)')Trim(TxtFunc),' n (n)log(n) integral exact, approx, error:'
            write(iout,fmti)'Integral n =',tb,n,tb,xlog,(tb,fint(k),k=1,3)
            Call ArrayPrint('x, df/dx, df/dx num., error, f:',x,df,f,fmtx=fmta)
            Call ArrayPrint('x, f,     f - interp.  , error, deriv.:',xi,fi,dfi,fmtx=fmta)
         endif
      End Subroutine Result_Print
   End Subroutine TestData
   ! -------------------------------------------
   subroutine TestInterpolation(which,x,w,A,xi,Li,q,Lp,f,df,fi,dfi,fint,sym)
      integer :: which,sym,k
      Real(float) :: x(sym:),w(sym:),A(sym:,sym:),xi(:),Li(:,sym:),Lp(:,sym:)
      Real(float) :: q(sym:,sym:),fint(:),fi(:,:),dfi(:,:),f(sym:),df(sym:,:)
      xi = -cos(pi*(real((/((k),k=0,ni-1)/),float)/real(ni-1))) ! calculate at chebyshev points
      if(shifted)xi = (1.0_float + xi)*0.5_float
      Li = Lcoeff(xi,OC)            ! interpolating polynomials at xi
      q = Lpoly(x)                  ! interpolating polynomial as monomials
      Lp = EvalPoly(xi,q)         ! evaluate interpolating polynomial
      ! ----- tests with exp(-ax^2)
      df = 0.0_float;  fi = 0.0_float;  fint = 0.0_float
      select case(which)
      case(1)  ! exponential function & integrals
      !* f(x) = exp(-5x^2) (integrals correct for all geometries)
         TxtFunc = 'exp(-5*x^2)'
         call ExpFunc(x,fint(1),f,df(:,1))         ! integral, f, df/dx at x coll. pts.
         call ExpFunc(xi,fint(1),fi(:,1),dfi(:,1)) ! integral, f, df/dx at x interpolation pts.
      case(2)  ! solution of example problem
      !* f(x) = 1 - cosh(5x)/cosh(5) (incorrect integrals for cylindrical & spherical)
         TxtFunc = 'cosh(5x)/cosh(5)'
         call Ex01Func(x,fint(1),f,df(:,1))         ! integral, f, df/dx at x coll. pts.
         call Ex01Func(xi,fint(1),fi(:,1),dfi(:,1)) ! integral, f, df/dx at x interpolation pts.
      case(3)  ! Runge function
      !* f(x) = 1/(1 + 25x^2) (incorrect integrals for cylindrical & spherical)
         TxtFunc = '1/(1 + 25x^2)'
         call RungeFunc(x,fint(1),f,df(:,1))         ! integral, f, df/dx at x coll. pts.
         call RungeFunc(xi,fint(1),fi(:,1),dfi(:,1)) ! integral, f, df/dx at x interpolation pts.
      end select
      df(:,2) = Matmul(A,f)                     ! approximate derivative at coll. pts
      fi(:,2) = Matmul(Li,f)                    ! interpolated function
      fint(2) = sum(w*f)                        ! approx. integral
      if(abs(fint(1)) > 0.0_float)fint(3) = abs(1.0_float-fint(2)/fint(1))
      df(:,3) = abs(df(:,1)-df(:,2))
      fi(:,3) = abs(fi(:,1)-fi(:,2))
      !where(abs(df(:,1)) > 0.0_float)df(:,3) = abs(1.0_float-df(:,2)/df(:,1))
      !where(abs(fi(:,1)) > 0.0_float)fi(:,3) = abs(1.0_float-fi(:,2)/fi(:,1))
   end subroutine TestInterpolation
   ! -------------------------------------------
   Function EvalPoly(xi,q) result(p)
      Real(float) :: xi(:),q(:,:),p(size(xi),size(q,1))
      Real(float) :: x(size(xi))
      integer :: i,j,n
      ! Evaluate monomial form of Lagrange polynomial
      n = size(q,1);  x = xi
      if(Symmetric)x = x*x
      do i=1,n
         p(:,i) = q(i,n)
         do j=n-1,1,-1
            p(:,i) = q(i,j) + p(:,i)*x
         enddo
      enddo
   End Function EvalPoly
   ! -------------------------------------------
   Integer Function QuadCheck(xq,wq,geom)  !  Check Quadrature accuracy
      Real(float), intent(in) :: xq(:),wq(:)
      integer, intent(in) :: geom
      Real(float) :: xx(size(xq)),x2(size(xq)),fint
      Integer :: k,m
      if(geom == Nonsymmetric)then
         m = 1;  x2 = xq      ! nonsymmetric
      else
         m = 2;  x2 = xq*xq   ! symmetric
      endif
      k = 0;  xx = 1.0_float
      do
         fint = Sum(xx*wq)
         if(abs(fint - 1.0_float/Real(m*k+igeom,float)) > 1.0e-12)exit
         xx = xx*x2;  k = k + 1
      enddo
      QuadCheck = k-1
   End Function QuadCheck
   ! -------------------------------------------
   Subroutine ExpFunc(xx,fint,f,df)                    ! exp(-5x^2) test function
      Use OrthogonalColloc, Only : Planar,Cylindrical,Spherical
      Real(float), intent(in) :: xx(:)                 ! to test interpolation
      Real(float), intent(out) :: fint,f(:),df(:)      ! integral,f(x),df/dx
      real(float) :: x(size(xx))
      logical :: shft
      Real(float), parameter :: aexp = -5.0_float, pi = 3.141592653589793_float
      ! ---------------------------------------
      !  x - value of x
      !  fint - integral of function
      !  f - f(x)
      !  df - df/dx
      ! ---------------------------------------
      x = xx
      shft = shifted
      shft = .true.  ! override, do from -1 to 1
      if(.not.shft)x = (xx + 1.0_float)*0.5_float
      f = exp(aexp*x*x)          ! function values
      df = (2.0_float)*aexp*x*f  ! derivative values
      select case(igeom)         ! integrals for various geometry
      case(Cylindrical) ! rdr term
         fint = (-0.5_float/aexp)*(1.0_float - exp(aexp))
      case(Spherical)   !r^2dr term
         fint = (0.25_float)*sqrt(-pi/aexp**3)*erf(sqrt(-aexp)) &
                  + (0.5_float/aexp)*exp(aexp)
      case default   ! planar dx
         fint = erf(sqrt(-aexp))*sqrt(-pi/aexp)*(0.5_float)
      End Select
      if(.not.shifted)fint = fint*2.0_float
      if(.not.shft)then
         fint = fint*2.0_float
         df = df*0.5_float
      endif
   End Subroutine ExpFunc
   ! -------------------------------------------
   Subroutine Ex01Func(xx,fint,f,df)
      Real(float), intent(in) :: xx(:)
      Real(float) :: fint,f(:),df(:)      ! integral,f(x),df/dx
      real(float) :: x(size(xx))
      Real(float), parameter :: parm = 5.0_float, div = 1.0_float/cosh(parm)
      Real(float) :: dfac = 1.0_float
      ! ---------------------------------------
      ! f,df & fint - function values, derivatives, integral
      ! f(x) = cosh(5*x)/cosh(5)
      ! solution to linear catalyst pellet problem
      ! this is not correct for cylindrical & spherical
      x = xx
      if(shifted .and. igeom<Planar)then
         x = x*2.0_float-1.0_float
         dfac = 2.0_float
      endif
      f = cosh(parm*x)*div             ! function values
      df = sinh(parm*x)*div*parm*dfac  ! derivative values
      fint = tanh(parm)/parm           ! integral (incorrect for cylindrical & spherical
      if(.not.shifted .and. igeom < Planar)fint = fint*2.0_float
   End Subroutine Ex01Func
   ! -------------------------------------------
   Subroutine RungeFunc(x,fint,f,df)                    ! exp(-5x^2) test function
      Real(float), intent(in) :: x(:)                 ! to test interpolation
      Real(float), intent(out) :: fint,f(:),df(:)      ! integral,f(x),df/dx
      Real(float), parameter :: aparm = 25.0_float, ap = sqrt(aparm)
      ! ---------------------------------------
      ! The Runge function f(x) = 1/(1 + 25x^2)
      !  x - value of x
      !  fint - integral of f correct only for planar
      !  f - f(x)
      !  df - df/dx
      ! ---------------------------------------
      f = 1.0_float/(1.0_float + aparm*x*x)
      df = f*f*x*(-2.0_float*aparm)
      fint = atan(ap)/ap
      if(.not.shifted)fint = fint*2.0_float
   End Subroutine RungeFunc
   ! -------------------------------------------
   Function MassFunc(xx) result(f)
      include 'defs.fi'
      Real(float), intent(in) :: xx(:)
      Real(float) :: x(size(xx)),f(size(xx))
      Real(float), parameter :: a0 = 0.20_float, a1 = (1.0_float - a0)*4.0_float
      ! ---------------------------------------
      ! function used for Mass Matrix calculation
      !  f(x) = 1 - x^2                if symmetric
      !  f(x) = 0.2 + 1.6(3 - 2x)x^2   if nonsymmetric
      ! ---------------------------------------
      x = xx
      if(Symmetric)then
         f = 1.0_float - x*x              ! Graetz problem
      else
        if(.not.shifted)x = (xx + 1.0_float)*0.5_float
        f = a0 + a1*(1.5_float - x)*x*x   ! BVP with cubic coef., 0.2 to 1.8
      endif
   End Function MassFunc
End Module TestModule
! ------------------------------------------------------------------------------
Program Test
   Use TestModule
   Use OrthogonalColloc, Only : Gauss,Lobatto,Chebyshev,RadauR,RadauL,General, &
                                Nonsymmetric,Planar,Cylindrical,Spherical,TxtType,TxtGeom
   Use Array_Print, Only : OpenFile
   include 'defs.fi'
   Integer :: n,typ
   Integer :: Nmin, Nmax,Ndel ! min and max collocation points
   Integer :: sym             ! coefficients start at 0 for nonsymmetric, 1 for symmetric
   Character :: fname*80,ch*1
   ! ------------------------------------------------------------------------------
   shifted = .false. ! set manually
   shifted = .true.
   sym = 0; typ = Gauss;  Nmin = 2;  Nmax = 2; igeom = Nonsymmetric
   write(*,'(a,4I2,a)')' Enter',Nonsymmetric,Planar,Cylindrical,Spherical,&
      ' for Nonsymmetric, Planar, Cylindrical, or Spherical geometry>'
   read(*,*)igeom
   sym = max(Nonsymmetric,min(igeom,1))
   if(sym == 1)then
      write(*,'(a)')' Enter g/l/c for Gauss/Lobatto/Chebyshev >'
      read(*,*)ch;  typ = index('glc',ch);  typ = Max(Gauss,Min(Chebyshev,typ))
      shifted = .true.
   else
      write(*,'(a,i2,a,i2,a)')' Enter g/l/c/r/x/e for Gauss/Lobatto/Chebyshev/Radau(R/L),Equal Space >'
      read(*,*)ch;  typ = index('eglcrx',ch)-1;  typ = Max(General,Min(RadauL,typ))
   endif
   write(*,'(a)')' Enter min,max,deta N>'
   read(*,*)Nmin,Nmax,Ndel
   Nmin = max(Nmin,0)
   Nmax = max(Nmin,Nmax)
   Ndel = max(Ndel,1)
   write(*,'(a)')'Test Mass Matrix T/F?'
   read(*,*)testm
   write(*,'(a)')'Test interpolation, integration, differentiation T/F?'
   read(*,*)testi
   write(*,'(A)')' Enter output file base name>'
   read(*,*)fname
   Call OpenFile(fname)
   write(iout,'(6a)')'Collocation Parameters for ',Trim(TxtGeom(igeom)),' Problems'
   do n=Nmin,Nmax,Ndel
      write(iout,'(/,3a,i3)')'Data for ',trim(TxtType(typ)),' Points, n =',n
      Call TestData(n,typ,sym)
   enddo
End Program Test

